>From acffac51cd72a98005ab7fd0e5f96f8d97b96c9f Mon Sep 17 00:00:00 2001
From: Roiy Shpaner <roiy@cs.umanitoba.ca>
Date: Tue, 15 Apr 2014 18:33:20 -0500
Subject: [PATCH 03/10] MED Support - File reader and writer modules

---
 include/vlc_meta.h        |   9 +
 modules/demux/Makefile.am |   3 +
 modules/demux/med.c       | 567 ++++++++++++++++++++++++++++++++++++++++++++++
 src/Makefile.am           |   1 +
 src/input/med_detect.c    | 402 ++++++++++++++++++++++++++++++++
 5 files changed, 982 insertions(+)
 create mode 100644 modules/demux/med.c
 create mode 100644 src/input/med_detect.c

diff --git a/include/vlc_meta.h b/include/vlc_meta.h
index bd7b319..3e8afa3 100644
--- a/include/vlc_meta.h
+++ b/include/vlc_meta.h
@@ -108,6 +108,15 @@ typedef struct meta_export_t
     const char *psz_file;
 } meta_export_t;
 
+typedef struct med_data_t
+{
+    VLC_COMMON_MEMBERS
+    input_thread_t *p_input;
+    const char *psz_file;
+    const char *psz_uri;
+    bool b_init;
+} med_data_t;
+
 VLC_API int input_item_WriteMeta(vlc_object_t *, input_item_t *);
 
 /* Setters for meta.
diff --git a/modules/demux/Makefile.am b/modules/demux/Makefile.am
index 4ab22af..6bf76d6 100644
--- a/modules/demux/Makefile.am
+++ b/modules/demux/Makefile.am
@@ -59,6 +59,9 @@ demux_LTLIBRARIES += libmjpeg_plugin.la
 libsubtitle_plugin_la_SOURCES = demux/subtitle.c
 demux_LTLIBRARIES += libsubtitle_plugin.la
 
+libmed_plugin_la_SOURCES = demux/med.c
+demux_LTLIBRARIES += libmed_plugin.la
+
 libty_plugin_la_SOURCES = demux/ty.c codec/cc.h
 demux_LTLIBRARIES += libty_plugin.la
 
diff --git a/modules/demux/med.c b/modules/demux/med.c
new file mode 100644
index 0000000..53cc8ce
--- /dev/null
+++ b/modules/demux/med.c
@@ -0,0 +1,567 @@
+/*****************************************************************************
+ * med.c: Media Enhacnement Data files (.med) processor. 
+ *****************************************************************************
+ * Copyright (C) 2014 VLC authors and VideoLAN
+ * $Id$
+ *
+ * Author: Roiy Shpaner <roiy@cs.umanitoba.ca>
+ *
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as published by
+ * the Free Software Foundation; either version 2.1 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+ * GNU Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
+ *****************************************************************************/
+
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
+#include <vlc_common.h>
+#include <vlc_plugin.h>
+#include <vlc_input.h>
+#include <vlc_memory.h>
+#include <vlc_fs.h>
+#include <vlc_url.h>
+#include <vlc_meta.h>
+
+#include <ctype.h>
+
+#include <vlc_demux.h>
+#include <vlc_charset.h>
+
+
+/*****************************************************************************
+ * Module descriptor
+
+ This module reads and writes the MED format file. Read events are saved in the input thread.
+ *****************************************************************************/
+static int  Open ( vlc_object_t *p_this );
+static void Close( vlc_object_t *p_this );
+static int  Control( demux_t *, int, va_list );
+static int  SaveMed ( vlc_object_t *p_this );
+
+vlc_module_begin ()
+    set_shortname( N_("MediaEnhancementData"))
+    set_description( N_("Media Enhancement Data file processor") )
+    set_capability( "demux", 0 )
+    set_category(CAT_ADVANCED)
+    set_callbacks( Open, Close )
+    add_shortcut( "med" )
+    add_submodule ()
+        set_capability( "med writer", 0 )
+        set_callbacks( SaveMed, NULL )
+        add_shortcut( "med writer" )
+vlc_module_end ()
+
+/*****************************************************************************
+ * Prototypes:
+ *****************************************************************************/
+
+typedef struct
+{
+    int     i_line_count;
+    int     i_line;
+    char    **line;
+} text_t;
+
+
+struct demux_sys_t
+{
+    int         i_type;
+    text_t      txt;
+    es_out_id_t *es;
+
+    int64_t     i_microsecperframe;
+};
+
+static int  TextLoad( text_t *, stream_t *s );
+static void TextUnload( text_t * );
+static int  ParseMED     ( text_t *txt, demux_t *p_demux, seeksegment_t *p_med, char***, int*, bool );
+static int  SaveEnhancementFile( input_thread_t *p_input, const char* origFileName );
+static int  AddLayer( char* layerName, char*** psz_layerNames , int* pi_layers);
+
+/*****************************************************************************
+ * Module initializer
+ *****************************************************************************/
+static int Open ( vlc_object_t *p_this )
+{
+    demux_t        *p_demux = (demux_t*)p_this;
+    demux_sys_t *p_sys;
+
+    p_demux->pf_control = Control;
+    p_demux->p_sys = p_sys = malloc( sizeof( demux_sys_t ) );
+    if( p_sys == NULL )
+        return VLC_ENOMEM;
+
+    seeksegment_t   *med_event = vlc_seeksegment_New();
+    int i_events = 0;
+
+    /* Probe file */
+    int     i_try;
+    char    *s = NULL;
+
+    msg_Dbg( p_demux, "Checking for med format" );
+    for( i_try = 0; i_try < 256; i_try++ )
+    {
+        if( ( s = stream_ReadLine( p_demux->s ) ) == NULL )
+        {
+            msg_Dbg( p_demux, "Problem with med file format, expecting med header" );
+            return VLC_EGENERIC;
+        }
+
+        else if( !strncasecmp( s, "med", 3  ) )
+        {
+            msg_Dbg( p_demux, "Detected med file start" );
+            free( s );
+            break;
+        }
+    }
+
+    msg_Dbg( p_demux, "loading med file..." );
+
+    /* Load the whole file */
+    TextLoad( &p_sys->txt, p_demux->s );
+    
+    /* get the layer names variables */
+    int* pi_layers;
+    char*** psz_layerNames;
+    input_Control( p_demux->p_input, INPUT_GET_LAYER_NAMES, &psz_layerNames, &pi_layers );
+
+
+    /* Parse it */
+    while ( ParseMED( &p_sys->txt, p_demux, med_event, psz_layerNames, pi_layers, 
+                     true ) == VLC_SUCCESS )
+    {
+        if ( input_Control( p_demux->p_input, INPUT_ADD_SEGMENT, med_event )
+            != VLC_SUCCESS )
+        {
+            msg_Err( p_demux, "Failed to add segment! #%d", i_events );
+            return VLC_EGENERIC;
+        }  
+        
+        vlc_seeksegment_Clear( med_event );
+        i_events++;
+    }
+
+    /* Unload */
+    TextUnload( &p_sys->txt );
+
+    msg_Dbg(p_demux, "loaded %d MED events", i_events );
+
+    vlc_seeksegment_Delete( med_event );
+
+    return VLC_SUCCESS;
+}
+
+/*****************************************************************************
+ * Control:
+ * Dummy control, this is a temporary module
+ *****************************************************************************/
+static int Control( demux_t *p_demux, int i_query, va_list args )
+{
+    VLC_UNUSED( p_demux );
+    VLC_UNUSED( args );
+    switch( i_query )
+    {
+        default:
+            return VLC_EGENERIC;
+    }
+}
+
+/*****************************************************************************
+ * Close: Close med demux
+ *****************************************************************************/
+static void Close( vlc_object_t *p_this )
+{
+    VLC_UNUSED( p_this );
+}
+
+static int TextLoad( text_t *txt, stream_t *s )
+{
+    int   i_line_max;
+
+    /* init txt */
+    i_line_max          = 500;
+    txt->i_line_count   = 0;
+    txt->i_line         = 0;
+    txt->line           = calloc( i_line_max, sizeof( char * ) );
+    if( !txt->line )
+        return VLC_ENOMEM;
+
+    /* load the complete file */
+    for( ;; )
+    {
+        char *psz = stream_ReadLine( s );
+
+        if( psz == NULL )
+            break;
+
+        txt->line[txt->i_line_count++] = psz;
+        if( txt->i_line_count >= i_line_max )
+        {
+            i_line_max += 100;
+            txt->line = realloc_or_free( txt->line, i_line_max * sizeof( char * ) );
+            if( !txt->line )
+                return VLC_ENOMEM;
+        }
+    }
+
+    if( txt->i_line_count <= 0 )
+    {
+        free( txt->line );
+        return VLC_EGENERIC;
+    }
+
+    return VLC_SUCCESS;
+}
+static void TextUnload( text_t *txt )
+{
+    int i;
+
+    for( i = 0; i < txt->i_line_count; i++ )
+    {
+        free( txt->line[i] );
+    }
+    free( txt->line );
+    txt->i_line       = 0;
+    txt->i_line_count = 0;
+}
+
+static char *TextGetLine( text_t *txt )
+{
+    if( txt->i_line >= txt->i_line_count )
+        return( NULL );
+
+    return txt->line[txt->i_line++];
+}
+
+/*****************************************************************************
+ * Specific MED function
+ *****************************************************************************/
+
+/* ParseMED
+ *  Format MED
+ *      MED(Case insensitive)
+ *
+ *      TargetName          //Link to media file name
+ *      
+ *      TargetHash          //Link to media file hash
+ *      
+ *      TargetSize          //Link to media file size
+ *
+ *      [[LayerName]n]      //This creates a unique event id. (LayerName is unique, so LayerName#n is as well)
+ *      h1:m1:s1,d1 --> h2:m2:s2,d2
+ *      Line1(optional)
+ *      Line2(optional)
+ *      ....
+ *      Line#(optional)
+ *      Empty line
+ *      Empty line
+ *
+ *      (More events in the format described above as needed)
+ *
+ *
+ *      MED(Case insensitive)
+ *      EOF
+ *
+ */
+static int ParseMED( text_t *txt, demux_t *p_demux, seeksegment_t *p_med, char ***psz_layerNames, int *pi_layers, 
+                                 bool b_replace_br )
+{
+    char    *psz_text;
+    const char *psz_fmt = "%d:%d:%d,%d --> %d:%d:%d,%d";
+
+    for( ;; )
+    {
+        const char *line = TextGetLine( txt );
+        int i_eventNum;
+        char* layerName;
+        char* parseChar;
+
+        if( !line )
+        {
+            return VLC_EGENERIC;
+        }
+
+        if ( line[0] == '[' && line[1] == '[' )     //[[MetaLayer]EventNumber]
+        {
+            if ( (parseChar = strchr( line , ']' )) )
+            {
+                if ( strchr( parseChar + 1 , ']' ) )    //there is a closing bracket
+                {
+                    if( sscanf( parseChar + 1, "%d", &i_eventNum ) == 1 )
+                    {
+                        int len = parseChar - line - 2;
+                        layerName = calloc( len, sizeof( char * ) );
+                        memcpy( layerName, line + 2 , len + 1);
+                        layerName[len] = '\0';
+
+                        int layerId = AddLayer( layerName, psz_layerNames , pi_layers );
+                
+                        //p_med->i_eventNum: i_eventNum is filled in the insertion, med data is ignored
+                        p_med->psz_layerName = layerName;
+                        p_med->i_layer_id = layerId;
+                        p_med->i_byte_offset = -1;      //currently unused
+                        p_med->i_byte_duration = -1;    //currently unused
+                        break;
+                    }
+
+                }
+            }
+        }
+    }
+
+
+    const char *line = TextGetLine( txt );
+    int h1, m1, s1, d1, h2, m2, s2, d2;
+
+    if( !line )
+        return VLC_EGENERIC;
+
+    if( sscanf( line, psz_fmt,      
+                &h1, &m1, &s1, &d1,
+                &h2, &m2, &s2, &d2 ) == 8 ) //read timestamp
+    {
+        p_med->i_time_offset = ( (int64_t)h1 * 3600*1000 +
+                                (int64_t)m1 * 60*1000 +
+                                (int64_t)s1 * 1000 +
+                                (int64_t)d1 ) * 1000;
+
+        int64_t i_stop  = ( (int64_t)h2 * 3600*1000 +
+                                (int64_t)m2 * 60*1000 +
+                                (int64_t)s2 * 1000 +
+                                (int64_t)d2 ) * 1000;
+
+        p_med->i_time_duration = i_stop - p_med->i_time_offset;
+
+        if( p_med->i_time_offset > i_stop )
+        {
+            msg_Err( p_demux, "Wrong timestamp format, read: %s", line );
+            return VLC_EGENERIC;
+        }
+            
+    }
+    else
+    {
+        msg_Err( p_demux, "Wrong format of file, expecting timestamp, read: %s", line );
+        return VLC_EGENERIC;
+    }
+
+
+    psz_text = strdup("");
+    if( !psz_text )
+        return VLC_ENOMEM;
+
+    for( ;; )    /* Now read text until double empty line */
+    {
+        const char *line = TextGetLine( txt );
+        int i_len;
+        int i_old;
+
+        i_len = line ? strlen( line ) : 0;
+        if( i_len <= 0 )
+        {
+            const char *newLine = TextGetLine( txt );
+
+            i_len = newLine ? strlen( newLine ) : 0;
+            if( i_len <= 0 )     //Double blank line - end of annotation
+            {
+                //Try to consume more blank lines
+                const char *extraLine;
+                do
+                {
+                    extraLine = TextGetLine( txt );
+                    i_len = extraLine ? strlen( extraLine ) : 0;
+                    if( i_len > 0 )
+                        txt->i_line--;  //End of blanks, go back
+                }
+                while ( extraLine && ( i_len <= 0 ) );
+
+                p_med->psz_text = psz_text;
+                return VLC_SUCCESS;
+            }
+
+            //one blank line, then annotation
+            //Apply the blank line before adding the annotation
+            i_old = strlen( psz_text );
+            psz_text = realloc_or_free( psz_text, i_old + 1 + 1 + 1 );
+            strcat( psz_text, "\n" );
+        }
+
+        i_old = strlen( psz_text );
+        psz_text = realloc_or_free( psz_text, i_old + i_len + 1 + 1 );
+        if( !psz_text )
+        {
+            return VLC_ENOMEM;
+        }
+
+        strcat( psz_text, line );
+        strcat( psz_text, "\n" );
+
+        /* replace [br] by \n */
+        if( b_replace_br )
+        {
+            char *p;
+
+            while( ( p = strstr( psz_text, "[br]" ) ) )
+            {
+                *p++ = '\n';
+                memmove( p, &p[3], strlen(&p[3])+1 );
+            }
+        }
+    }
+}
+
+/* Save file with current events and annotations, defaults to media file name with med extension */ 
+static int SaveEnhancementFile( input_thread_t *p_input, const char* origFileName )
+{
+    if ( !origFileName )
+    {
+        msg_Err( p_input, "No filename given to save med data" );
+        return VLC_EGENERIC;
+    }
+
+    FILE *pFile = vlc_fopen( origFileName, "wt" );
+    if( pFile == NULL )
+    {
+        msg_Err( p_input, "could not create .med file %s (%m)",  origFileName );
+        return VLC_EGENERIC;
+    }
+
+    /* Get events and annotations */
+    seeksegment_t **seeksegments;
+    int i_seeksegments;
+    if ( input_Control( p_input, INPUT_GET_SEGMENTS, &seeksegments, &i_seeksegments, NULL ) != VLC_SUCCESS )
+    {
+        msg_Err( p_input, "Could not get MED events!" );
+        fclose(pFile);
+        return VLC_EGENERIC;
+    }
+
+    /* Write to file */
+    fprintf(pFile, "med\n\n");
+
+    for ( int i = 0; i < i_seeksegments ; i++ )
+    {
+        fprintf(pFile, "[[%s]%d]\n", seeksegments[i]->psz_layerName, seeksegments[i]->i_eventNum);
+
+        long startTime = seeksegments[i]->i_time_offset / 1000;
+        long endTime = ( seeksegments[i]->i_time_offset + seeksegments[i]->i_time_duration ) / 1000;
+        int ms = startTime % 1000;
+        int endMs = endTime % 1000;
+        
+        startTime /= 1000;
+        endTime /= 1000;
+        int seconds = startTime % 60;
+        int endSeconds = endTime % 60;
+
+        startTime /= 60;
+        endTime /= 60;
+        int minutes = startTime % 60;
+        int endMinutes = endTime % 60;
+
+        startTime /= 60;
+        endTime /= 60;
+        int hours = startTime;
+        int endHours = endTime;
+
+        fprintf(pFile, "%02d:%02d:%02d,%03d --> %02d:%02d:%02d,%03d\n", hours, minutes, seconds, ms, endHours, endMinutes, endSeconds, endMs );
+        
+        //Remove double spaced lines - it ends annotation
+        char* newLine = strstr( seeksegments[i]->psz_text, "\n\n" );
+        while ( newLine )
+        {
+            while ( newLine[1] == '\n' )
+            {
+                newLine[1] = ' ';   //replace \n characters with spaces - no more than one blank line allowed
+                newLine++;
+            }
+
+            newLine = strstr( newLine, "\n\n" );
+        }
+
+        fprintf(pFile, "%s", seeksegments[i]->psz_text );
+
+        int count = 3;
+        int pos = strlen( seeksegments[i]->psz_text ) - 1;
+        if  ( ( pos < 0 ) || ( seeksegments[i]->psz_text[pos] == '\n' ) )
+                count = 2;
+
+        for ( ; count > 0 ; count-- )  //Add necessary newlines
+            fprintf(pFile, "\n" );
+    }
+
+    fprintf(pFile, "med");
+    fclose(pFile);
+
+    vlc_seeksegment_Delete_arr( seeksegments, i_seeksegments );
+ 
+    return VLC_SUCCESS;
+}
+
+/* Adds layer of events to array if it did not exist before */
+static int AddLayer( char* layerName, char*** psz_layerNames , int* pi_layers)
+{
+    for ( int i = 0 ; i < *pi_layers ; i++ )
+    {
+        if ( strnicmp( (*psz_layerNames)[i], layerName, 255 ) == 0 )  /* Layer found, no need to add */
+            return i;
+    }
+
+    (*pi_layers)++;
+    *psz_layerNames = realloc_or_free( *psz_layerNames, (*pi_layers) * sizeof(char*) );   //Add layer to array
+    (*psz_layerNames)[(*pi_layers)-1] = strdup( layerName );
+
+    return (*pi_layers)-1;
+}
+
+/* Initializesa new empty MED file */
+static int InitEnhancementFile( input_thread_t *p_input, char* filename )
+{
+    FILE *pFile = vlc_fopen( filename, "wt" );
+
+    if( pFile == NULL )
+    {
+        msg_Err( p_input, "could not create .med file: %s ",  filename );
+        return VLC_EGENERIC;
+    }
+
+    fprintf(pFile, "med\n\n\nmed");
+    fclose(pFile);
+
+    return VLC_SUCCESS;
+}
+
+static int SaveMed ( vlc_object_t *p_this )
+{
+    med_data_t *p_med = (med_data_t *)p_this;
+    char* filename;
+
+    if ( !p_med->psz_file )  /* Default med file name - media filename with med extension */
+    {
+        filename = make_path( p_med->psz_uri );
+        char* dot = strrchr( filename, '.' ); /* Replace extension with .med */
+        *dot  = '\0';
+        filename = realloc( filename, strlen( filename ) + 4 + 1 );
+        if ( !filename ) 
+            return VLC_ENOMEM;
+
+        strcat( filename, ".med" );           
+    }
+    else
+        filename = p_med->psz_file;
+    
+    if ( p_med->b_init )
+        return InitEnhancementFile( p_med->p_input, filename );
+    else
+        return SaveEnhancementFile( p_med->p_input, filename );
+}
diff --git a/src/Makefile.am b/src/Makefile.am
index 842fc73..0c0a079 100644
--- a/src/Makefile.am
+++ b/src/Makefile.am
@@ -388,6 +388,7 @@ SOURCES_libvlc_common = \
 	input/stream_filter.c \
 	input/stream_memory.c \
 	input/subtitles.c \
+	input/med_detect.c \
 	input/var.c \
 	video_output/chrono.h \
 	video_output/control.c \
diff --git a/src/input/med_detect.c b/src/input/med_detect.c
new file mode 100644
index 0000000..0572b3a
--- /dev/null
+++ b/src/input/med_detect.c
@@ -0,0 +1,402 @@
+/*****************************************************************************
+ * med_detect.c : Media Enhancement Data files detection
+ *****************************************************************************
+ * Copyright (C) 2013 VLC authors and VideoLAN
+ * $Id$
+ *
+ * Authors: Roiy Shpaner <roiy@cs.umaniotba.ca>
+ *
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as published by
+ * the Free Software Foundation; either version 2.1 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+ * GNU Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
+ *****************************************************************************/
+
+/**
+ *  \file
+ *  This file contains functions to detect MED(.med) files.
+ */
+
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
+#include <ctype.h> /* isalnum() */
+#ifdef HAVE_UNISTD_H
+#   include <unistd.h>
+#endif
+#include <sys/stat.h>
+
+#include <vlc_common.h>
+#include <vlc_fs.h>
+#include <vlc_url.h>
+
+#include "input_internal.h"
+
+/**
+ * We are not going to autodetect more med files than this.
+ */
+#define MAX_MED_FILES 128
+
+
+static void strcpy_trim( char *d, const char *s )
+{
+    unsigned char c;
+
+    /* skip leading whitespace */
+    while( ((c = *s) != '\0') && !isalnum(c) )
+    {
+        s++;
+    }
+    for(;;)
+    {
+        /* copy word */
+        while( ((c = *s) != '\0') && isalnum(c) )
+        {
+            *d = tolower(c);
+            s++; d++;
+        }
+        if( *s == 0 ) break;
+        /* trim excess whitespace */
+        while( ((c = *s) != '\0') && !isalnum(c) )
+        {
+            s++;
+        }
+        if( *s == 0 ) break;
+        *d++ = ' ';
+    }
+    *d = 0;
+}
+
+static void strcpy_strip_ext( char *d, const char *s )
+{
+    unsigned char c;
+
+    const char *tmp = strrchr(s, '.');
+    if( !tmp )
+    {
+        strcpy(d, s);
+        return;
+    }
+    else
+        strlcpy(d, s, tmp - s + 1 );
+    while( (c = *d) != '\0' )
+    {
+        *d = tolower(c);
+        d++;
+    }
+}
+
+static void strcpy_get_ext( char *d, const char *s )
+{
+    const char *tmp = strrchr(s, '.');
+    if( !tmp )
+        strcpy(d, "");
+    else
+        strcpy( d, tmp + 1 );
+}
+
+static int whiteonly( const char *s )
+{
+    unsigned char c;
+
+    while( (c = *s) != '\0' )
+    {
+        if( isalnum( c ) )
+            return 0;
+        s++;
+    }
+    return 1;
+}
+
+enum
+{
+    SUB_PRIORITY_NONE        = 0,
+    SUB_PRIORITY_MATCH_NONE  = 1,
+    SUB_PRIORITY_MATCH_RIGHT = 2,
+    SUB_PRIORITY_MATCH_LEFT  = 3,
+    SUB_PRIORITY_MATCH_ALL   = 4,
+};
+typedef struct
+{
+    int priority;
+    char *psz_fname;
+    char *psz_ext;
+} vlc_medfn_t;
+
+static int compare_med_priority( const void *a, const void *b )
+{
+    const vlc_medfn_t *p0 = a;
+    const vlc_medfn_t *p1 = b;
+
+    if( p0->priority > p1->priority )
+        return -1;
+
+    if( p0->priority < p1->priority )
+        return 1;
+
+#ifdef HAVE_STRCOLL
+    return strcoll( p0->psz_fname, p1->psz_fname);
+#else
+    return strcmp( p0->psz_fname, p1->psz_fname);
+#endif
+}
+
+/*
+ * Check if a file ends with a med extension
+ */
+int med_Filter( const char *psz_dir_content )
+{
+    const char *tmp = strrchr( psz_dir_content, '.');
+
+    if( !tmp )
+        return 0;
+    tmp++;
+
+    if( strcasecmp( "med", tmp ) == 0 )
+        return 1;
+    return 0;
+}
+
+
+/**
+ * Convert a list of paths separated by ',' to a char**
+ */
+static char **paths_to_list( const char *psz_dir, char *psz_path )
+{
+    unsigned int i, k, i_nb_subdirs;
+    char **subdirs; /* list of subdirectories to look in */
+    char *psz_parser = psz_path;
+
+    if( !psz_dir || !psz_path )
+        return NULL;
+
+    for( k = 0, i_nb_subdirs = 1; psz_path[k] != '\0'; k++ )
+    {
+        if( psz_path[k] == ',' )
+            i_nb_subdirs++;
+    }
+
+    subdirs = calloc( i_nb_subdirs + 1, sizeof(char*) );
+    if( !subdirs )
+        return NULL;
+
+    for( i = 0; psz_parser && *psz_parser != '\0' ; )
+    {
+        char *psz_subdir = psz_parser;
+        psz_parser = strchr( psz_subdir, ',' );
+        if( psz_parser )
+        {
+            *psz_parser++ = '\0';
+            while( *psz_parser == ' ' )
+                psz_parser++;
+        }
+        if( *psz_subdir == '\0' )
+            continue;
+
+        if( asprintf( &subdirs[i++], "%s%s",
+                  psz_subdir[0] == '.' ? psz_dir : "",
+                  psz_subdir ) == -1 )
+            break;
+    }
+    subdirs[i] = NULL;
+
+    return subdirs;
+}
+
+
+/**
+ * Detect med files.
+ *
+ * When called this function will split up the psz_name string into a
+ * directory, filename and extension. It then opens the directory
+ * in which the file resides and tries to find possible matches of
+ * med files.
+ *
+ * \ingroup Demux
+ * \param p_this the calling \ref input_thread_t
+ * \param psz_path a list of subdirectories (separated by a ',') to look in.
+ * \param psz_name the complete filename to base the search on.
+ * \return a NULL terminated array of filenames with detected possible med files.
+ * The array contains max MAX_MED_FILES items and you need to free it after use.
+ */
+char **med_Detect( input_thread_t *p_this, char *psz_path,
+                         const char *psz_name_org )
+{
+    int j, i_result2, i_med_count, i_fname_len;
+    char *f_fname_noext = NULL, *f_fname_trim = NULL;
+
+    char **subdirs; /* list of subdirectories to look in */
+
+    vlc_medfn_t *result = NULL; /* unsorted results */
+    char **result2; /* sorted results */
+
+    if( !psz_name_org )
+        return NULL;
+
+    char *psz_fname = make_path( psz_name_org );
+    if( !psz_fname )
+        return NULL;
+
+    /* extract filename & dirname from psz_fname */
+    char *f_dir = strdup( psz_fname );
+    if( f_dir == NULL )
+    {
+        free( psz_fname );
+        return NULL;
+    }
+
+    const char *f_fname = strrchr( psz_fname, DIR_SEP_CHAR );
+    if( !f_fname )
+    {
+        free( f_dir );
+        free( psz_fname );
+        return NULL;
+    }
+    f_fname++; /* Skip the '/' */
+    f_dir[f_fname - psz_fname] = 0; /* keep dir separator in f_dir */
+
+    i_fname_len = strlen( f_fname );
+
+    f_fname_noext = malloc(i_fname_len + 1);
+    f_fname_trim = malloc(i_fname_len + 1 );
+    if( !f_fname_noext || !f_fname_trim )
+    {
+        free( f_dir );
+        free( f_fname_noext );
+        free( f_fname_trim );
+        free( psz_fname );
+        return NULL;
+    }
+
+    strcpy_strip_ext( f_fname_noext, f_fname );
+    strcpy_trim( f_fname_trim, f_fname_noext );
+
+    result = calloc( MAX_MED_FILES+1, sizeof(vlc_medfn_t) ); /* We check it later (simplify code) */
+    subdirs = paths_to_list( f_dir, psz_path );
+    for( j = -1, i_med_count = 0; (j == -1) || ( j >= 0 && subdirs != NULL && subdirs[j] != NULL ); j++ )
+    {
+        const char *psz_dir = (j < 0) ? f_dir : subdirs[j];
+        if( psz_dir == NULL || ( j >= 0 && !strcmp( psz_dir, f_dir ) ) )
+            continue;
+
+        /* parse psz_src dir */
+        DIR *dir = vlc_opendir( psz_dir );
+        if( dir == NULL )
+            continue;
+
+        msg_Dbg( p_this, "looking for a med file in %s", psz_dir );
+
+        char *psz_name;
+        while( (psz_name = vlc_readdir( dir )) && i_med_count < MAX_MED_FILES )
+        {
+            if( psz_name[0] == '.' || !med_Filter( psz_name ) )
+                continue;
+
+            char tmp_fname_noext[strlen( psz_name ) + 1];
+            char tmp_fname_trim[strlen( psz_name ) + 1];
+            char tmp_fname_ext[strlen( psz_name ) + 1];
+            const char *tmp;
+            int i_prio = SUB_PRIORITY_NONE;
+
+            /* retrieve various parts of the filename */
+            strcpy_strip_ext( tmp_fname_noext, psz_name );
+            strcpy_get_ext( tmp_fname_ext, psz_name );
+            strcpy_trim( tmp_fname_trim, tmp_fname_noext );
+
+            if( !strcmp( tmp_fname_trim, f_fname_trim ) )
+            {
+                /* matches the movie name exactly */
+                i_prio = SUB_PRIORITY_MATCH_ALL;
+            }
+            else if( (tmp = strstr( tmp_fname_trim, f_fname_trim )) )
+            {
+                /* contains the movie name */
+                tmp += strlen( f_fname_trim );
+                if( whiteonly( tmp ) )
+                {
+                    /* chars in front of the movie name */
+                    i_prio = SUB_PRIORITY_MATCH_RIGHT;
+                }
+                else
+                {
+                    /* chars after (and possibly in front of)
+                     * the movie name */
+                    i_prio = SUB_PRIORITY_MATCH_LEFT;
+                }
+            }
+            else if( j == -1 )
+            {
+                /* doesn't contain the movie name, prefer files in f_dir over subdirs */
+                i_prio = SUB_PRIORITY_MATCH_NONE;
+            }
+
+            char psz_fullpath[strlen( psz_dir ) + strlen( psz_name ) + 2];
+            struct stat st;
+
+            sprintf( psz_fullpath, "%s"DIR_SEP"%s", psz_dir, psz_name );
+            if( !strcmp( psz_fullpath, psz_fname ) )
+                continue;
+
+            if( !vlc_stat( psz_fullpath, &st ) && S_ISREG( st.st_mode ) && result && ( i_prio == SUB_PRIORITY_MATCH_ALL ) )
+            {
+                msg_Dbg( p_this, "autodetected med file: %s with priority %d", psz_fullpath, i_prio );
+                result[i_med_count].priority = i_prio;
+                result[i_med_count].psz_fname = strdup( psz_fullpath );
+                result[i_med_count].psz_ext = strdup(tmp_fname_ext);
+                i_med_count++;
+            }
+        }
+        closedir( dir );
+    }
+    if( subdirs )
+    {
+        for( j = 0; subdirs[j]; j++ )
+            free( subdirs[j] );
+        free( subdirs );
+    }
+    free( f_dir );
+    free( f_fname_trim );
+    free( f_fname_noext );
+    free( psz_fname );
+
+    if( !result )
+        return NULL;
+
+    qsort( result, i_med_count, sizeof(vlc_medfn_t), compare_med_priority );
+
+    result2 = calloc( i_med_count + 1, sizeof(char*) );
+
+    for( j = 0, i_result2 = 0; j < i_med_count && result2 != NULL; j++ )
+    {
+        bool b_reject = false;
+
+        if( !result[j].psz_fname || !result[j].psz_ext ) /* memory out */
+            break;
+
+        /* */
+        if( !b_reject )
+            result2[i_result2++] = strdup( result[j].psz_fname );
+
+        msg_Dbg( p_this, "Found result: %s", result2[i_result2-1] );
+    }
+
+    for( j = 0; j < i_med_count; j++ )
+    {
+        free( result[j].psz_fname );
+        free( result[j].psz_ext );
+    }
+    free( result );
+
+    return result2;
+}
+
-- 
1.8.4.msysgit.0

